// Copyright Alexander Blach © 2016

(function() {
    // this calls the -[WKScriptMessageHandler userContentController:didReceiveScriptMessage:] method
    function callMessageHandler(functionName, message, url, line, column, callerFunctionName) {
        var message = {
            'functionName': functionName,
            'message': message,
            'url': url,
            'line': line,
            'column': column,
            'callerFunctionName': callerFunctionName
        };
        window.webkit.messageHandlers.consoleLog.postMessage(message);
    }

    // parse the url and location of a line in the stack trace
    // see https://github.com/WebKit/webkit/blob/15c73993b1a656a237dc2d89a31e26b38d53416a/Source/WebInspectorUI/UserInterface/Models/StackTrace.js#L102
    function parseLocation(locationString) {
        var result = { url: "", line: 0, column: 0 };
        var locationRegEx = /(.+?)(?::(\d+)(?::(\d+))?)?$/;
        var matched = locationString.match(locationRegEx);

        if (!matched) {
            return result;
        }

        result.url = matched[1];

        if (matched[2]) {
            result.line = parseInt(matched[2]);
        }

        if (matched[3]) {
            result.column = parseInt(matched[3]);
        }

        return result;
    }

    function parseErrorStack(stack) {
        // uncomment for debugging
        // callMessageHandler("log", stack);

        var callerLine = stack.split("\n")[1];
        var result = { url: "", line: 0, column: 0, functionName: "" };

        if (callerLine) {
            // example string:
            // test@http://localhost:52320/External/7F7F656E-4A55-4F71-BDD3-03E061C47716/test.html:8:15
            // adapted from https://github.com/WebKit/webkit/blob/15c73993b1a656a237dc2d89a31e26b38d53416a/Source/WebInspectorUI/UserInterface/Models/StackTrace.js#L76

            var atIndex = callerLine.indexOf("@");

            if (atIndex !== -1) {
                result.functionName = callerLine.slice(0, atIndex);
                Object.assign(result, parseLocation(callerLine.slice(atIndex + 1)));
            } else if (callerLine.includes("/")) {
                Object.assign(result, parseLocation(callerLine));
            } else {
                result.functionName = callerLine;
            }
        }

        return result;
    }

    // see https://github.com/moll/json-stringify-safe
    function safeStringify(obj, replacer, spaces, cycleReplacer) {
        return JSON.stringify(obj, serializer(replacer, cycleReplacer), spaces);
    }

    function serializer(replacer, cycleReplacer) {
        var stack = [],
            keys = [];

        if (cycleReplacer == null) cycleReplacer = function(key, value) {
            if (stack[0] === value) return "[Circular ~]";
            return "[Circular ~." + keys.slice(0, stack.indexOf(value)).join(".") + "]";
        }

        return function(key, value) {
            if (stack.length > 0) {
                var thisPos = stack.indexOf(this);
                ~thisPos ? stack.splice(thisPos + 1) : stack.push(this);
                ~thisPos ? keys.splice(thisPos, Infinity, key) : keys.push(key);
                if (~stack.indexOf(value)) {
                    value = cycleReplacer.call(this, key, value);
                }
            } else { stack.push(value); }

            return replacer == null ? value : replacer.call(this, key, value);
        }
    }

    // replace methods of console object with our own function that calls
    // the WKScriptMessageHandler in addition to the original function
    for (var name of ["log", "info", "warn", "error", "assert", "clear", "debug"]) {
        // use an anonymous function here so that the func variable below gets its own scope
        (function(name) {
            // get original function
            var func = console[name];
            if (typeof func == 'function') {
                // replace with our own function
                console[name] = function() {
                    // try to get caller information from stack
                    var caller = parseErrorStack((new Error).stack);

                    // convert arguments to JavaScript array
                    var args = Array.prototype.slice.call(arguments);

                    // call Objective-C code
                    callMessageHandler(func.name, JSON.parse(safeStringify(args)), caller.url, caller.line, caller.column, caller.functionName);

                    // call original method
                    return func.apply(console, arguments);
                }
            }
        }(name));
    }

    var errorName = 'error';
    window.addEventListener(errorName, function(event) {
        callMessageHandler(errorName, event.message, event.filename, event.lineno, event.colno);
    });
})()