'use strict';

hterm.defaultStorage = new lib.Storage.Memory();
var _preferences = new hterm.PreferenceManager('textastic');
var _terminal;


/////////////////////////////////////////
// customize some of hterm's methods

hterm.Terminal.prototype.setWindowTitle = function(title) {
    window.document.title = title;

    // Textastic
    terminalCallMessageHandler('terminalWindowTitleChange', { title });
};

// this method is called in hterm.VT.prototype.setDECMode
hterm.Terminal.prototype.setApplicationCursorMode = function(state) {
    terminalCallMessageHandler('terminalSetApplicationCursorMode', { state });
}
hterm.Terminal.prototype.setApplicationKeypadMode = function(state) {
    terminalCallMessageHandler('terminalSetApplicationKeypadMode', { state });
}
hterm.Terminal.prototype.playBellAudio = function() {
    terminalCallMessageHandler('terminalPlayBellAudio', {});
}

let lastScrollHeight;
function syncScrollHeight() {
    const scrollHeight = parseFloat(_terminal.scrollPort_.scrollArea_.style.height);
    if (scrollHeight != lastScrollHeight) {
        terminalCallMessageHandler('scrollHeightChange', { scrollHeight });
        lastScrollHeight = scrollHeight;
    }
}

const origSyncScrollHeight = hterm.ScrollPort.prototype.syncScrollHeight;
hterm.ScrollPort.prototype.syncScrollHeight = function() {
    origSyncScrollHeight.call(this);
    syncScrollHeight();
};

let lastScrollTop;
// this function is called in the scroll event listener added in onTerminalReady()
function syncScrollTop() {
    // delay to give the scroll view a change to update its content size before scrolling
    setTimeout(function() {
        const scrollTop = _terminal.scrollPort_.screen_.scrollTop;
        if (scrollTop != lastScrollTop) {
            terminalCallMessageHandler('scrollTopChange', { scrollTop });
            lastScrollTop = scrollTop;
        }
    }, 5);    
}

/////////////////////////////////////////



// this calls the -[WKScriptMessageHandler userContentController:didReceiveScriptMessage:] method
function terminalCallMessageHandler(functionName, data) {
    var message = {
        'functionName': functionName,
        'data': data
    };
    window.webkit.messageHandlers.terminal.postMessage(message);
}

function terminalSetPreference(key, value) {
    _preferences.set(key, value);
}

function loadFont(family, fileName, descriptors) {
    var fontFace = new FontFace(family, 'url("' + fileName + '")', descriptors);
    fontFace.load();
    document.fonts.add(fontFace);
}

function loadFonts() {
    // load fonts
    var family;

    family = "Anonymous Pro";
    loadFont(family, "Anonymous Pro.ttf");
    loadFont(family, "Anonymous Pro B.ttf", { weight: "bold" });
    loadFont(family, "Anonymous Pro I.ttf", { style: "italic" });
    loadFont(family, "Anonymous Pro BI.ttf", { weight: "bold", style: "italic" });

    family = "CamingoCode";
    loadFont(family, "CamingoCode-Regular.ttf");
    loadFont(family, "CamingoCode-Bold.ttf", { weight: "bold" });
    loadFont(family, "CamingoCode-Italic.ttf", { style: "italic" });
    loadFont(family, "CamingoCode-BoldItalic.ttf", { weight: "bold", style: "italic" });

    family = "DejaVu Sans Mono";
    loadFont(family, "DejaVuSansMono.ttf");
    loadFont(family, "DejaVuSansMono-Bold.ttf", { weight: "bold" });
    loadFont(family, "DejaVuSansMono-Oblique.ttf", { style: "oblique" });
    loadFont(family, "DejaVuSansMono-BoldOblique.ttf", { weight: "bold", style: "oblique" });

    family = "Fira Code";
    loadFont(family, "FiraCode-Regular.otf");
    loadFont(family, "FiraCode-Bold.otf", { weight: "bold" });

    family = "Fira Mono";
    loadFont(family, "FiraMono-Regular.otf");
    loadFont(family, "FiraMono-Bold.otf", { weight: "bold" });

    family = "Inconsolata";
    loadFont(family, "Inconsolata-Regular.ttf");
    loadFont(family, "Inconsolata-Bold.ttf", { weight: "bold" });
    
    family = "JetBrains Mono";
    loadFont(family, "JetBrainsMono-Regular.ttf");
    loadFont(family, "JetBrainsMono-Bold.ttf", { weight: "bold" });
    loadFont(family, "JetBrainsMono-Italic.ttf", { style: "italic" });
    loadFont(family, "JetBrainsMono-Bold-Italic.ttf", { weight: "bold", style: "italic" });

    family = "Luxi Mono";
    loadFont(family, "luximr.ttf");
    loadFont(family, "luximb.ttf", { weight: "bold" });
    loadFont(family, "luximri.ttf", { style: "oblique" });
    loadFont(family, "luximbi.ttf", { weight: "bold", style: "oblique" });

    family = "monofur";
    loadFont(family, "monof55.ttf");
    loadFont(family, "monof56.ttf", { style: "italic" });

    family = "Source Code Pro";
    loadFont(family, "SourceCodePro-Regular.otf");
    loadFont(family, "SourceCodePro-Bold.otf", { weight: "bold" });
    loadFont(family, "SourceCodePro-It.otf", { style: "italic" });
    loadFont(family, "SourceCodePro-BoldIt.otf", { weight: "bold", style: "italic" });

    document.fonts.ready.then(function() {
        if (_terminal) {
            // re-measure character size
            _terminal.setFontSize(0);
        }
    });
}

function terminalInit(phone) {
    loadFonts();

    lib.init(function() {
        // set preferences
        terminalSetPreference('cursor-blink', true);
        terminalSetPreference('copy-on-select', false);
        terminalSetPreference('use-default-window-copy', true);
        terminalSetPreference('mouse-right-click-paste', false);
        terminalSetPreference('clear-selection-after-copy', false);
        terminalSetPreference('receive-encoding', 'raw');
        terminalSetPreference('audible-bell-sound', '');
        terminalSetPreference('enable-resize-status', true);
        terminalSetPreference('scroll-on-output', true);

        //terminalSetPreference('font-size', phone ? 12 : 14);

        var terminal = new hterm.Terminal('textastic');
        _terminal = terminal;

        terminal.onTerminalReady = function() {
            terminal.uninstallKeyboard();
            terminal.setCursorPosition(0, 0);
            terminal.setCursorVisible(true);

            terminal.scrollPort_.screen_.addEventListener('scroll', syncScrollTop);

            document.body.style.backgroundColor = _preferences.get('background-color');

            terminal.io.onTerminalResize = function(width, height) {
                terminalCallMessageHandler('terminalResize', { width, height });
            };

            terminalCallMessageHandler('terminalReady', {
                width: terminal.screenSize.width,
                height: terminal.screenSize.height
            });

            terminalCallMessageHandler('terminalSetApplicationCursorMode', { 'state': terminal.keyboard.applicationCursor });
            terminalCallMessageHandler('terminalSetApplicationKeypadMode', { 'state': terminal.keyboard.applicationKeypad });
        };

        terminal.decorate(document.getElementById('terminal'));

        // handle selection change
        document.addEventListener('selectionchange', function() {
            var text = '';
            var firstRangeRect = null;

            var selection = document.getSelection();
            if (selection && !selection.isCollapsed && selection.rangeCount > 0) {
                text = selection.toString();
                var rect = selection.getRangeAt(0).getBoundingClientRect();
                firstRangeRect = { x: rect.x, y: rect.y, width: rect.width, height: rect.height };
            }

            terminalCallMessageHandler('selectionChange', { text, firstRangeRect });
        });

    });
}

function terminalWrite(string) {
    _terminal.interpret(string);
}

function terminalFocus(focused) {
    _terminal.onFocusChange_(focused);
}

function terminalClearSelection() {
    document.getSelection().removeAllRanges();
}

function terminalIsApplicationCursorMode() {
    return _terminal.keyboard.applicationCursor;
}

function terminalSetApplicationCursorMode(state) {
    _terminal.keyboard.applicationCursor = state;
}

function terminalSetApplicationKeypadMode(state) {
    _terminal.keyboard.applicationKeypad = state;
}

function terminalGetCursorRect() {
    var rect = _terminal.cursorNode_.getBoundingClientRect();
    return { x: rect.x, y: rect.y, width: rect.width, height: rect.height };
}

function terminalSetFontSize(fontSize) {
    terminalSetPreference('font-size', fontSize);
    // we need to call setFontSize directly because setting the preference does not fire the notification when
    // we set the font size to the default value (15)
    if (_terminal) {
        _terminal.setFontSize(fontSize);
    }
}

function terminalSetFontFamily(family) {
    terminalSetPreference('font-family', family + ', Menlo, monospace');

    setTimeout(function() {
        if (_terminal) {
            // re-measure character size, shouldn't be necessary, but without this, the character
            // size is wrong after changing the font
            _terminal.setFontSize(0);
        }
    }, 10);
}

function terminalSetThemeName(name) {
    if (name == "Solarized Dark") {
        terminalSetPreference('color-palette-overrides', ["#002831", "#d11c24", "#738a05", "#a57706", "#2176c7", "#c61c6f", "#259286", "#eae3cb", "#001e27", "#bd3613", "#475b62", "#536870", "#708284", "#5956ba", "#819090", "#fcf4dc"]);
        terminalSetPreference('foreground-color', "#708284");
        terminalSetPreference('background-color', "#001e27");
        terminalSetPreference('cursor-color', 'rgba(112,130,132,0.5)');
    } else if (name == "Solarized Light") {
        terminalSetPreference('color-palette-overrides', ["#002831", "#d11c24", "#738a05", "#a57706", "#2176c7", "#c61c6f", "#259286", "#eae3cb", "#001e27", "#bd3613", "#475b62", "#536870", "#708284", "#5956ba", "#819090", "#fcf4dc"]);
        terminalSetPreference('foreground-color', "#536870");
        terminalSetPreference('background-color', "#fcf4dc");
        terminalSetPreference('cursor-color', 'rgba(83,104,112,0.5)');
    } else if (name == "Espresso Libre") {
        terminalSetPreference('color-palette-overrides', ["#000000", "#cc0000", "#1a921c", "#f0e53a", "#0066ff", "#c5656b", "#06989a", "#d3d7cf", "#555753", "#ef2929", "#9aff87", "#fffb5c", "#43a8ed", "#ff818a", "#34e2e2", "#eeeeec"]);
        terminalSetPreference('foreground-color', "#b8a898");
        terminalSetPreference('background-color', "#2a211c");
        terminalSetPreference('cursor-color', 'rgba(255,255,255,0.5)');
    } else if (name == "Monokai Soda") {
        terminalSetPreference('color-palette-overrides', ["#1a1a1a", "#f4005f", "#98e024", "#fa8419", "#9d65ff", "#f4005f", "#58d1eb", "#c4c5b5", "#625e4c", "#f4005f", "#98e024", "#e0d561", "#9d65ff", "#f4005f", "#58d1eb", "#f6f6ef"]);
        terminalSetPreference('foreground-color', "#c4c5b5");
        terminalSetPreference('background-color', "#1a1a1a");
        terminalSetPreference('cursor-color', 'rgba(246,247,236,0.5)');
    } else if (name == "Light") {
        terminalSetPreference('color-palette-overrides', []);
        terminalSetPreference('foreground-color', "#000000");
        terminalSetPreference('background-color', "#ffffff");
        terminalSetPreference('cursor-color', 'rgba(0, 131, 246, 0.6)');
    } else if (name == "Obsidian") {
        terminalSetPreference('color-palette-overrides', ["#000000", "#a60001", "#00bb00", "#fecd22", "#3a9bdb", "#bb00bb", "#00bbbb", "#bbbbbb", "#555555", "#ff0003", "#93c863", "#fef874", "#a1d7ff", "#ff55ff", "#55ffff", "#ffffff"]);
        terminalSetPreference('foreground-color', "#cdcdcd");
        terminalSetPreference('background-color', "#283033");
        terminalSetPreference('cursor-color', 'rgba(192,202,208,0.5)');
    } else {
        // default theme
        terminalSetPreference('color-palette-overrides', []);
        terminalSetPreference('foreground-color', '#f0f0f0');
        terminalSetPreference('background-color', '#101010');
        terminalSetPreference('cursor-color', 'rgba(71, 169, 255, 0.6)');
    }
    const color = _preferences.get('background-color');
    document.body.style.backgroundColor = color;
    terminalCallMessageHandler('backgroundColorChange', { color });
}

function terminalSetScrollTop(scrollTop) {
    _terminal.scrollPort_.screen_.scrollTop = scrollTop;
    // read calculated value
    lastScrollTop = _terminal.scrollPort_.screen_.scrollTop;
}

function terminalScrollToBottom() {
    _terminal.scrollEnd();
}
